/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* rc-package-info.c
 * Copyright (C) 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <gnome-xml/xmlmemory.h>
#include <gnome-xml/parser.h>

#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "rc-world.h"
#include "rc-package-info.h"
#include "rc-debman-general.h"
#include "rc-debug.h"
#include "rc-util.h"
#include "xml-util.h"
#include "rc-package-update.h"
#include "rc-xml.h"

typedef enum _DepType DepType;

enum _DepType {
    DEP_TYPE_REQUIRES,
    DEP_TYPE_PROVIDES,
    DEP_TYPE_CONFLICTS,
    DEP_TYPE_SUGGESTS,
    DEP_TYPE_RECOMMENDS
};

static guint rc_channel_parse_helix (RCChannel *, gchar *);
static guint rc_channel_parse_debian (RCChannel *, gchar *);
static guint rc_channel_parse_redhat (RCChannel *, gchar *);

guint
rc_channel_parse (RCChannel *channel, gchar *tbuf, int compressed_length)
{
    gchar *buf;
    GByteArray *byte_array = NULL;
    guint ret = 0;

    g_assert (tbuf);

    if (compressed_length) {
        if (rc_uncompress_memory (tbuf, compressed_length, &byte_array)) {
            g_warning ("Uncompression failed");
            return 1;
        }

        buf = byte_array->data;
    } else {
        buf = tbuf;
    }

    switch (channel->type) {
    case RC_CHANNEL_TYPE_HELIX:
        ret = rc_channel_parse_helix (channel, buf);
        break;
    case RC_CHANNEL_TYPE_DEBIAN:
        ret = rc_channel_parse_debian (channel, buf);
        break;
    case RC_CHANNEL_TYPE_REDHAT:
        ret = rc_channel_parse_redhat (channel, buf);
        break;
    default:
        g_error ("Trying to parse unknown channel type!");
        break;
    }

    if (byte_array) {
        g_byte_array_free (byte_array, TRUE);
    }

    return (ret);
}

static guint
rc_channel_parse_helix (RCChannel *channel, gchar *buf)
{
    xmlDoc *xml_doc;
    xmlNode *root;
    gboolean err = FALSE;

    g_assert (buf);

    /*
     * For some reason, either because of differences in the RCWorld or
     * because we're using libxml1 instead of libxml2, the SAX parser is
     * much slower than the DOM parser, so use it by default.
     */

    if (getenv ("RC_SAX_XML")) {
        RCPackageSAXContext *ctx;
        RCPackageSList *packages;
        RCWorld *world = rc_get_world();
        guint count;

        ctx = rc_package_sax_context_new(channel);
        rc_package_sax_context_parse_chunk(ctx, buf, strlen(buf));
        packages = rc_package_sax_context_done(ctx);

        count = g_slist_length(packages);

        rc_world_freeze(world);
        rc_world_add_packages_from_slist(world, packages);
        rc_world_thaw(world);

        g_hash_table_freeze (channel->packages);
        g_hash_table_freeze (channel->dep_table);
        g_hash_table_freeze (channel->dep_name_table);
        
        while (packages) {
            const RCPackageDepSList *prov_iter;
            RCPackage *package;
            
            package = (RCPackage *) packages->data;

            for (prov_iter = package->provides; prov_iter;
                 prov_iter = prov_iter->next)
            {
                const RCPackageDep *dep = (RCPackageDep *)(prov_iter->data);
                
                rc_hash_table_slist_insert_unique (channel->dep_table,
                                                   (gpointer) &dep->spec, package, NULL);
                rc_hash_table_slist_insert_unique (channel->dep_name_table,
                                                   dep->spec.name, package, NULL);
            }
            
            g_hash_table_insert (channel->packages,
                                 (gpointer) package->spec.name,
                                 (gpointer) package);
            
            packages = packages->next;
        }
        
        g_hash_table_thaw (channel->packages);
        g_hash_table_thaw (channel->dep_table);
        g_hash_table_thaw (channel->dep_name_table);
        
        g_slist_free (packages);

        if (count == 0)
            err = TRUE;
    }
    else {
        xml_doc = xmlParseMemory (buf, strlen (buf));
        
        if (!xml_doc) {
            return -1;
        }
        
        root = xmlDocGetRootElement (xml_doc);
        err = rc_xml_node_to_channel (channel, root);
        
        xmlFreeDoc (xml_doc);
    }

    return err;
}

static void debian_packages_helper (gchar *mbuf, RCPackage *pkg, gchar *url_prefix);

static guint
rc_channel_parse_debian (RCChannel *rcc, char *buf)
{
    /* Note that buf is NOT ours -- we can't munge it, and strdup'ing the
     * whole thing would be ufly. So, we let packages_query_helper handle
     */
    GSList *ps;
    RCWorld *world = rc_get_world ();

    gchar *b, *ob;

    g_hash_table_freeze (rcc->packages);
    g_hash_table_freeze (rcc->dep_table);
    g_hash_table_freeze (rcc->dep_name_table);

    /* Keep looking for a "Packages: " */
    ob = buf;
    while ((b = strstr (ob, "Package: ")) != NULL) {
        RCPackage *p;
        if (b != buf && b[-1] != '\r' && b[-1] != '\n') {
            ob = b+1;
            continue;
        }
        p = g_new0 (RCPackage, 1);
        /* All debian packages "have" epochs */
        p->spec.has_epoch = 1;
        debian_packages_helper (b, p, rcc->file_path);
        ((RCPackageUpdate *)(p->history->data))->package = p;
        ob = b+1;

        p->channel = rcc;

        ps = p->provides;
        while (ps) {
            RCPackageDep *dep = (RCPackageDep *) ps->data;
            rc_hash_table_slist_insert_unique (rcc->dep_table, &dep->spec, p, NULL);
            rc_hash_table_slist_insert_unique (rcc->dep_name_table, dep->spec.name, p, NULL);
            ps = ps->next;
        }

        rc_world_add_package (world, p);
        g_hash_table_insert (rcc->packages, p->spec.name, p);
        rc_hash_table_slist_insert_unique (rcc->dep_table, &p->spec, p, NULL);
        rc_hash_table_slist_insert_unique (rcc->dep_name_table, p->spec.name, p, NULL);
    }

    g_hash_table_thaw (rcc->packages);
    g_hash_table_thaw (rcc->dep_table);
    g_hash_table_thaw (rcc->dep_name_table);

    return 0;
}

/*
 * This code needs to be kept super fast
 */

static void
debian_packages_helper (gchar *mbuf, RCPackage *pkg, gchar *url_prefix)
{
    gchar *ibuf = mbuf;
    gchar *buf, *p;
    RCPackageUpdate *up = NULL;
    gboolean desc = FALSE;
    int ilen;

    up = rc_package_update_new ();
    /* All debian packages "have" epochs */
    up->spec.has_epoch = 1;

    while (1) {
        /* FIXME -- handle all empty line with spaces */
        if (*ibuf == '\0' || *ibuf == '\n') break;

        buf = strchr (ibuf, '\n');
        if (buf == NULL) break; /* something bizzare  happened */

        ilen = buf - ibuf;      /* the length of the line */
        buf = ibuf;             /* start of buf again */
        buf[ilen] = '\0';       /* we probably shouldn't modify the string in-place, but
                                 * we restore it at the end */
        ibuf += ilen + 1;

        /* Downcase the sucker - I hate debian - it's SUPPOSED to be nicely
         * capitalized, but.. */
        p = buf;
        if (*p != ' ')
            while (*p != ':') *p++ = tolower(*p);

        if (!strncmp (buf, "package: ", strlen ("package: "))) {
            pkg->spec.name = g_strndup (buf + strlen ("package: "),
                                        ilen - strlen ("package: "));
        } else if (!strncmp (buf, "installed-size: ", strlen ("installed-size: "))) {
            up->installed_size = strtoul (buf +
                                          strlen ("installed-size: "),
                                          NULL, 10) * 1024;
        } else if (!strncmp (buf, "size: ", strlen ("size: "))) {
            up->package_size = strtoul(buf + strlen("size: "), NULL, 10);
        } else if (!strncmp (buf, "description: ", strlen ("description: "))) {
            pkg->summary = g_strndup (buf + strlen ("description: "),
                                      ilen - strlen ("description: "));
            desc = TRUE;
        } else if (*buf == ' ' && desc) {
            if (ilen == 2 && buf[1] == '.') {
                /* It's just a newline */
                if (pkg->description) {
                    int l = strlen (pkg->description);
                    gchar *tmp = g_malloc (l + 2);
                    strcpy (tmp, pkg->description);
                    tmp[l] = '\n';
                    tmp[l+1] = '\0';
                    g_free (pkg->description);
                    pkg->description = tmp;
                } else {
                    pkg->description = g_strndup ("\n", 1);
                }
            } else {
                /* More than a newline */
                if (pkg->description) {
                    int l = strlen (pkg->description);
                    gchar *tmp = g_malloc (l + ilen + 1);
                    strcpy (tmp, pkg->description);
                    tmp[l] = '\n';
                    strncpy (tmp + l + 1, buf+1, ilen-1);
                    tmp[l + ilen] = '\0';
                    g_free (pkg->description);
                    pkg->description = tmp;
                } else {
                    pkg->description = g_strndup (buf+1, ilen-1);
                }
            }
        } else if (!strncmp (buf, "version: ", strlen ("version: "))) {
            rc_debman_parse_version (buf + strlen ("Version: "),
                                     &pkg->spec.epoch, &pkg->spec.version,
                                     &pkg->spec.release);
        } else if (!strncmp (buf, "section: ", strlen ("section: "))) {
            gchar **tokens = g_strsplit (buf + strlen ("section: "), "/", -1);
            gint i = 0;
            gchar *sec;
            while (tokens[i+1] != NULL) i++;
            sec = g_strchomp (tokens[i]);
            p = sec;
            while (*p) *p++ = tolower(*p);
            pkg->section = rc_debman_section_to_package_section (sec);
#if 0
            if (!strcmp (sec, "admin")) {
                pkg->section = RC_SECTION_SYSTEM;
            } else if (!strcmp (sec, "base")) {
                pkg->section = RC_SECTION_SYSTEM;
            } else if (!strcmp (sec, "comm")) {
                pkg->section = RC_SECTION_INTERNET;
            } else if (!strcmp (sec, "devel")) {
                pkg->section = RC_SECTION_DEVEL;
            } else if (!strcmp (sec, "doc")) {
                pkg->section = RC_SECTION_DOC;
            } else if (!strcmp (sec, "editors")) {
                pkg->section = RC_SECTION_UTIL;
            } else if (!strcmp (sec, "electronics")) {
                pkg->section = RC_SECTION_MISC;
            } else if (!strcmp (sec, "games")) {
                pkg->section = RC_SECTION_GAME;
            } else if (!strcmp (sec, "graphics")) {
                pkg->section = RC_SECTION_IMAGING;
            } else if (!strcmp (sec, "hamradio")) {
                pkg->section = RC_SECTION_MISC;
            } else if (!strcmp (sec, "interpreters")) {
                pkg->section = RC_SECTION_DEVELUTIL;
            } else if (!strcmp (sec, "libs")) {
                pkg->section = RC_SECTION_LIBRARY;
            } else if (!strcmp (sec, "mail")) {
                pkg->section = RC_SECTION_PIM;
            } else if (!strcmp (sec, "math")) {
                pkg->section = RC_SECTION_MISC;
            } else if (!strcmp (sec, "misc")) {
                pkg->section = RC_SECTION_MISC;
            } else if (!strcmp (sec, "net")) {
                pkg->section = RC_SECTION_INTERNET;
            } else if (!strcmp (sec, "news")) {
                pkg->section = RC_SECTION_INTERNET;
            } else if (!strcmp (sec, "oldlibs")) {
                pkg->section = RC_SECTION_LIBRARY;
            } else if (!strcmp (sec, "otherosfs")) {
                pkg->section = RC_SECTION_MISC;
            } else if (!strcmp (sec, "science")) {
                pkg->section = RC_SECTION_MISC;
            } else if (!strcmp (sec, "shells")) {
                pkg->section = RC_SECTION_SYSTEM;
            } else if (!strcmp (sec, "sound")) {
                pkg->section = RC_SECTION_MULTIMEDIA;
            } else if (!strcmp (sec, "tex")) {
                pkg->section = RC_SECTION_MISC;
            } else if (!strcmp (sec, "text")) {
                pkg->section = RC_SECTION_UTIL;
            } else if (!strcmp (sec, "utils")) {
                pkg->section = RC_SECTION_UTIL;
            } else if (!strcmp (sec, "web")) {
                pkg->section = RC_SECTION_INTERNET;
            } else if (!strcmp (sec, "x11")) {
                pkg->section = RC_SECTION_XAPP;
            } else {
                pkg->section = RC_SECTION_MISC;
            }
#endif
            g_strfreev (tokens);
        } else if (!strncmp (buf, "depends: ", strlen ("depends: "))) {
            pkg->requires = g_slist_concat (
                pkg->requires,
                rc_debman_fill_depends (buf + strlen ("depends: ")));
        } else if (!strncmp (buf, "recommends: ", strlen ("recommends: "))) {
            pkg->recommends = g_slist_concat (
                pkg->recommends,
                rc_debman_fill_depends (buf + strlen ("recommends: ")));
        } else if (!strncmp (buf, "suggests: ", strlen ("suggests: "))) {
            pkg->suggests = g_slist_concat (
                pkg->suggests,
                rc_debman_fill_depends (buf + strlen ("suggests: ")));
        } else if (!strncmp (buf, "pre-depends: ", strlen ("pre-depends: "))) {
            RCPackageDepSList *tmp, *iter;

            tmp = rc_debman_fill_depends (buf + strlen ("pre-depends: "));
            for (iter = tmp; iter; iter = iter->next) {
                RCPackageDep *dep = (RCPackageDep *)(iter->data);

                dep->pre = TRUE;
            }
            pkg->requires = g_slist_concat (
                pkg->requires, tmp);
        } else if (!strncmp (buf, "conflicts: ", strlen ("conflicts: "))) {
            pkg->conflicts = g_slist_concat (
                pkg->conflicts,
                rc_debman_fill_depends (buf + strlen ("conflicts: ")));
        } else if (!strncmp (buf, "provides: ", strlen ("provides: "))) {
            pkg->provides = g_slist_concat (
                pkg->provides,
                rc_debman_fill_depends (buf + strlen ("provides: ")));

        } else if (!strncmp (buf, "filename: ", strlen ("filename: "))) {
            /* Build a new update with just this version */
            if (url_prefix) {
                up->package_url = g_strconcat (url_prefix, "/",
                                               buf + strlen ("filename: "),
                                               NULL);
            } else {
                up->package_url = g_strndup (buf + strlen ("filename: "),
                                             ilen - strlen ("filename: "));
            }
        } else if (!strncmp (buf, "md5sum: ", strlen ("md5sum: "))) {
            up->md5sum = g_strndup (buf + strlen ("md5sum: "),
                                    ilen - strlen ("md5sum: "));
        }

        buf[ilen] = '\n';
    }

    up->importance = RC_IMPORTANCE_SUGGESTED;
    up->description = g_strdup ("Upstream Debian release");
    rc_package_spec_copy (&up->spec, &pkg->spec);
    pkg->history = g_slist_append (pkg->history, up);

    /* Make sure to provide myself, for the dep code! */
    pkg->provides =
        g_slist_append (pkg->provides, rc_package_dep_new_from_spec
                        (&pkg->spec,
                         RC_RELATION_EQUAL));
}


static guint
rc_channel_parse_redhat (RCChannel *rcc, char *buf)
{
    g_error ("No redhat channel support!");
    return 0;
}

/* Helper functions */

RCPackage *
pkginfo_find_package_with_constraint (RCChannel *ch,
                                      gchar *name,
                                      RCPackageDep *constraint,
                                      gboolean is_and)

{
    RCPackage *rp, *candidate = NULL;
    RCPackageUpdate *upi;

    g_return_val_if_fail (ch != NULL, NULL);
    
#if 0
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "pfpwc: name: %s constr: ", name);
    rc_package_dep_dump (constraint);
#endif

    rp = g_hash_table_lookup (ch->packages, name);

    if (rp) {
#if DEBUG > 10
        rc_debug (RC_DEBUG_LEVEL_DEBUG, "   found %s in hash...\n", rp->spec.name);
#endif
        if (constraint) {
            RCPackageDep *dep;

            g_assert(rp->history);
            g_assert(rp->history->data);
            
            upi = (RCPackageUpdate *) rp->history->data;

            dep = rc_package_dep_new_from_spec (&(upi->spec),
                                                RC_RELATION_EQUAL);

            if (rc_package_dep_verify_relation
                (constraint, dep))
                {
                    candidate = rp;
                }

            rc_package_dep_free (dep);
        }
        else {
            candidate = rp;
        }
        if (candidate)
            return candidate;
#if DEBUG > 10
        if (constraint)
            rc_debug (RC_DEBUG_LEVEL_DEBUG, "   failed constraint test.\n");
        else
            rc_debug (RC_DEBUG_LEVEL_DEBUG, "   didn't exist!\n");
#endif
    }

    return NULL;
}

RCPackage *
pkginfo_find_package (RCChannel *chs,
                      gchar *name)
{
    return pkginfo_find_package_with_constraint (chs, name, NULL, FALSE);
}
