/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * gui-priority.c: Priority Updates functions and utilities
 *
 * Copyright 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <fcntl.h>

#include <gnome.h>
#include <libsoup/soup.h>
#include <uuid/uuid.h>

#include "gui.h"
#include "gui-html.h"
#include "gui-prefs.h"
#include "gui-priority.h"
#include "gui-transfer.h"
#include "gui-uri.h"
#include "gui-util.h"
#include "mirror.h"

#define DEFAULT_PRIORITY_HOST_ROOT "https://web-bos-ma-us.rc.ximian.com"
#define ACTIVATION_CODE_URL        "https://activation.rc.ximian.com/reg-code.php"

/* There is a bug in older versions of the client which messes up on
 * activation URLs passed back by the server that aren't in the root of the
 * document root.  This means that older clients will have a hard time with
 * activation on the standard 443 port.  In order to maintain backward
 * compatibility, we now pass the following version code as part of the
 * query string, so that the server can pass back the old activation URL on
 * port 8000 for older clients, and the newer 443 port for us.
 */
#define ACTIVATION_VERSION 2

#define HTTP_RESPONSE_SUCCESSFUL(x)   (x >= 200 && x < 300)

typedef struct {
    GtkHTML *html;
    char *method;
    char *cookie;
    char *secret;
} SubmissionInfo;

typedef enum {
    MANIFEST_UPDATE,
    MANIFEST_INSTALL,
    MANIFEST_REMOVE
} ManifestAction;

gboolean priority_set = FALSE;

static void
set_priority_mode(gboolean priority)
{
    appinfo->priority = priority;
    if (priority) {
        gtk_widget_show(GET_WIDGET("priority_pixmap"));
        gtk_label_set_text(
            GTK_LABEL(GET_WIDGET("priority_label")),
            _("Premium services enabled"));
    }
    else {
        gtk_widget_hide(GET_WIDGET("priority_pixmap"));
        gtk_label_set_text(GTK_LABEL(GET_WIDGET("priority_label")), "");
    }
} /* set_priority_mode */

void
rc_create_cookie(void)
{
    uuid_t uuid;
    char *out;
    FILE *f;

    /* User does not want a cookie generated on this machine. */
    if (g_file_exists("/etc/ximian/no-cookie"))
        return;

    /* A cookie already exists for this system. */
    if (g_file_exists("/etc/ximian/mcookie"))
        return;

    out = g_malloc0(37);
    uuid_generate_random(uuid);
    uuid_unparse(uuid, out);

    rc_mkdir("/etc/ximian", 0755);
    f = fopen("/etc/ximian/mcookie", "w");
    if (!f) {
        g_free(out);
        g_warning("Unable to create /etc/ximian/mcookie");
    }
    else {
        fwrite(out, 37, 1, f);
        fflush(f);
        fclose(f);
    }
} /* rc_create_cookie */

static char *
rc_get_cookie(gboolean create)
{
	char *secret;
    int fd;

    secret = g_malloc0(37);

    if (!g_file_exists("/etc/ximian/mcookie") &&
        !g_file_exists("/etc/ximian/no-cookie")) {
        if (create)
            rc_create_cookie();
        else {
            g_free(secret);
            return NULL;
        }
    }

    fd = open("/etc/ximian/mcookie", O_RDONLY);
    read(fd, secret, 36);
    close(fd);

    secret [36] = '\0';

    return secret;
} /* rc_get_cookie */

static char *
rc_get_secret(gboolean create)
{
    char *secret;
    uuid_t uuid;

    secret = g_malloc0(37);

    if (g_file_exists("/etc/ximian/partnernet")) {
        int fd;
        int b;

        fd = open("/etc/ximian/partnernet", O_RDONLY);
        b = read(fd, secret, 36); // don't read the trailing \n
        secret[36] = '\0';
        close(fd);

        if (b >= 36)
            return secret;
    }

    if (!create) {
        g_free(secret);
        return NULL;
    }

    uuid_generate_random(uuid);
    uuid_unparse(uuid, secret);

    return secret;
} /* rc_get_secret */

/*
 * old_pkg is NULL if action is not UPDATE
 */
static xmlNode *
manifest_xml_node(int cid, RCPackage *new_pkg, RCPackage *old_pkg,
                  ManifestAction action, const char *url)
{
    xmlNode *node, *pkgnode;
    char *tmp = NULL;

    node = xmlNewNode(NULL, "manifest");

    if (cid) {
        tmp = g_strdup_printf("%d", cid);
        xmlNewTextChild(node, NULL, "cid", tmp);
        g_free(tmp);
    }

    switch (action) {
    case MANIFEST_UPDATE:
        tmp = "update";
        break;
    case MANIFEST_INSTALL:
        tmp = "install";
        break;
    case MANIFEST_REMOVE:
        tmp = "remove";
        break;
    default:
        g_assert_not_reached();
        break;
    }
    xmlNewTextChild(node, NULL, "action", tmp);

    /* write info on the new package */
    pkgnode = xmlNewChild (node, NULL, "package", NULL);
    xmlNewTextChild(pkgnode, NULL, "name", new_pkg->spec.name);
    tmp = g_strdup_printf("%d", new_pkg->spec.epoch);
    xmlNewTextChild(pkgnode, NULL, "epoch", tmp);
    g_free(tmp);
    xmlNewTextChild(pkgnode, NULL, "version", new_pkg->spec.version);
    xmlNewTextChild(pkgnode, NULL, "release", new_pkg->spec.release);
    if (action != MANIFEST_REMOVE) {
        RCPackageUpdate *upd = rc_package_get_latest_update(new_pkg);

        tmp = g_strdup_printf("%d", upd->package_size);
        xmlNewTextChild(pkgnode, NULL, "size", tmp);
        g_free(tmp);

        tmp = g_strdup_printf("%d", upd->hid);
        xmlNewTextChild(pkgnode, NULL, "hid", tmp);
        g_free(tmp);

        if (new_pkg->channel) {
            tmp = g_strdup_printf("%d", new_pkg->channel->id);
            xmlNewTextChild(pkgnode, NULL, "channel_id", tmp);
            g_free(tmp);
        }

        if (url)
            xmlNewTextChild(pkgnode, NULL, "url", url);
    }

    /* write info on the old package, if any */
    if (old_pkg) {
        pkgnode = xmlNewChild (node, NULL, "oldpackage", NULL);
        tmp = g_strdup_printf("%d", old_pkg->spec.epoch);
        xmlNewTextChild(pkgnode, NULL, "epoch", tmp);
        g_free(tmp);
        xmlNewTextChild(pkgnode, NULL, "version", old_pkg->spec.version);
        xmlNewTextChild(pkgnode, NULL, "release", old_pkg->spec.release);
    }

    return node;
} /* manifest_xml_node */

static xmlChar *
transaction_xml(RCPackageSList *install_packages, 
                RCPackageSList *remove_packages,
                int *bytes)
{
    xmlDoc *doc;
    char *mid = rc_get_cookie(TRUE);
    time_t curtime = time(NULL);
    char *tmp;
    RCPackageSList *i;
    xmlChar *xml_string;

    doc = xmlNewDoc("1.0");
    doc->root = xmlNewDocNode(doc, NULL, "transaction", NULL);

    xmlNewTextChild(doc->root, NULL, "mid", mid);
    
    tmp = g_strdup_printf("%ld", curtime);
    xmlNewTextChild(doc->root, NULL, "start_time", tmp);
    g_free(tmp);

    xmlNewTextChild(
        doc->root, NULL, "distro",
        rc_distro_get_target()
        );

    i = install_packages;
    while (i) {
        RCPackage *p = i->data;
        RCPackageUpdate *u = rc_package_get_latest_update(p);
        RCPackage *sys_pkg;
        ManifestAction action;
        xmlNode *n;

        sys_pkg = rc_package_new();
        sys_pkg->spec.name = g_strdup(p->spec.name);
        sys_pkg = rc_packman_query(appinfo->pman, sys_pkg);

        if (sys_pkg->installed)
            action = MANIFEST_UPDATE;
        else
            action = MANIFEST_INSTALL;

        n = manifest_xml_node(
            p->channel->id, p, sys_pkg->installed ? sys_pkg : NULL,
            action, u->package_url);

        rc_package_free(sys_pkg);

        xmlAddChild(doc->root, n);

        i = i->next;
    }

    i = remove_packages;
    while (i) {
        RCPackage *p = i->data;
        xmlNode *n;

        n = manifest_xml_node(
            0, p, NULL, MANIFEST_REMOVE, NULL);

        xmlAddChild(doc->root, n);

        i = i->next;
    }

    xmlDocDumpMemory(doc, &xml_string, bytes);
            
    return xml_string;
} /* transaction_xml */

static void
transaction_sent_cb(SoupMessage *message, SoupErrorCode err, gpointer data)
{
    char **tid = data;
    const char *tid_string = soup_message_get_response_header(
        message, "X-TID");

    if (err)
        g_warning("An error was returned from the logging server: %d", err);

    if (tid_string) {
        *tid = g_strdup(tid_string);
    }
    else
        g_warning("No X-TID response");

    /* Not a g_free() because this is an xmlChar * */
    if (message->request.body)
        free(message->request.body);
} /* transaction_sent_cb */

void
rc_send_transaction(RCPackageSList *install_packages, 
                    RCPackageSList *remove_packages, char **tid)
{
    xmlChar *xml_string;
    int bytes;
    char *url;
    SoupUri *soup_uri;
    SoupContext *context;
    SoupMessage *message;

    xml_string = transaction_xml(install_packages, remove_packages, &bytes);

    url = rc_gui_get_mirrored_url("log.php", NULL);
    soup_uri = rc_gui_priority_authorized_url(url);

    if (!soup_uri) {
        rc_debug(RC_DEBUG_LEVEL_WARNING,
                 __FUNCTION__ "Invalid URL: %s\n", url);
        return;
    }

    context = soup_context_from_uri(soup_uri);
    g_free(url);
    message = soup_message_new_full(
        context, NULL, SOUP_BUFFER_USER_OWNED, xml_string, bytes);
    if (gui_prefs_get_http10_enabled())
        soup_message_set_http_version(message, SOUP_HTTP_1_0);
    soup_context_unref(context);
    soup_uri_free(soup_uri);

    soup_message_queue(message, transaction_sent_cb, tid);
} /* rc_send_transaction */

static xmlChar *
success_xml(char *tid, gboolean successful, char *message, int *bytes)
{
    xmlDoc *doc;
    xmlChar *xml_string;
    gchar *tmp;
    time_t curtime = time(NULL);
    char *mid = rc_get_cookie(TRUE);

    doc = xmlNewDoc("1.0");
    doc->root = xmlNewDocNode(doc, NULL, "transaction_end", NULL);

    xmlNewTextChild(doc->root, NULL, "mid", mid);
    xmlNewTextChild(doc->root, NULL, "tid", tid);
    xmlNewTextChild(doc->root, NULL, "success", successful ? "1" : "0");

    tmp = g_strdup_printf("%ld", curtime);
    xmlNewTextChild(doc->root, NULL, "end_time", tmp);
    g_free(tmp);

    if (message)
        xmlNewTextChild(doc->root, NULL, "message", message);

    xmlDocDumpMemory(doc, &xml_string, bytes);
            
    return xml_string;
} /* success_xml */

static void
success_sent_cb(SoupMessage *message, SoupErrorCode err, gpointer data)
{
    char *tid = data;

    g_free(tid);

    /* Not a g_free() because it's an xmlChar * */
    if (message->request.body)
        free(message->request.body);
} /* transaction_sent_cb */

void
rc_send_success(char *tid, gboolean successful, char *msg)
{
    xmlChar *xml_string;
    int bytes;
    char *url;
    SoupUri *soup_uri;
    SoupContext *context;
    SoupMessage *message;

    if (!tid) {
        /* There's no tid available, so we can't really send a success code */
        return;
    }

    xml_string = success_xml(tid, successful, msg, &bytes);

    url = rc_gui_get_mirrored_url("log.php", NULL);
    soup_uri = rc_gui_priority_authorized_url(url);

    if (!soup_uri) {
        rc_debug(RC_DEBUG_LEVEL_WARNING,
                 __FUNCTION__ "Invalid URL: %s\n", url);
        return;
    }

    context = soup_context_from_uri(soup_uri);
    g_free(url);
    message = soup_message_new_full(
        context, NULL, SOUP_BUFFER_USER_OWNED, xml_string, bytes);
    if (gui_prefs_get_http10_enabled())
        soup_message_set_http_version(message, SOUP_HTTP_1_0);
    soup_context_unref(context);
    soup_uri_free(soup_uri);

    soup_message_queue(message, success_sent_cb, tid);
} /* rc_send_success */

static gboolean
refresh_channels(gpointer data)
{
    rc_gui_refresh_all_channels();

    return FALSE;
} /* refresh_channels */

static void
enter_priority_mode(GtkButton *button)
{
    gui_prefs_set_priority(TRUE);
    gui_prefs_sync_priority();
    set_priority_mode(TRUE);
    gtk_widget_set_sensitive(GET_WIDGET("priority_frame"), TRUE);
    gtk_idle_add(refresh_channels, NULL);
#if 0
    /* Commented out because this results in a soup call from inside a
     *  soup call, and as it turns out, soup isn't very reentrant.
     *
     * 13:22 <@joe> hmm
     * 13:22 <@joe> soup is not very reentrant
     * 13:22 <@jpr> joe: not with the static buffers
     * 13:23 <@joe> ok...
     * 13:23 <@joe> nothing a little idle handler won't fix
     */
    
    rc_gui_refresh_all_channels();
#endif
} /* enter_priority_mode */
  
static void
submit_done(SoupMessage *message, SoupErrorCode err, gpointer data)
{
    SubmissionInfo *info = data;
    GtkHTML *html = info->html;
    GtkHTMLStream *stream;
    const char *response;

    rc_dump_transfer_info(message);
    rc_debug(RC_DEBUG_LEVEL_INFO,
             "[%p]: Soup Error code: %d\n", message, err);

    if (err) {
        char *url;
        char *e;

        g_warning("Submission error: %d.\nHTTP Error code: %d %s\n",
                  err, message->response_code, message->response_phrase);

        url = soup_uri_to_string(soup_context_get_uri(message->context), TRUE);
        e = g_strdup_printf(
            _("<p>Red Carpet encountered a network error while trying to "
              "submit your activation information.  Please contact Ximian "
              "support for assistance.</p>"
              "<p>Details:<br>"
              "URL: %s<br>"
              "HTTP Response: %d %s<br>"
              "Soup Error: %d</p>"),
            url, message->response_code, message->response_phrase, err);
        rc_gui_error_box(e, RC_ERROR_ABORT);
        g_free(url);
        g_free(e);

        rc_gui_show_executive_summary();

        goto SUBMIT_CLEANUP;
    }

    response = soup_message_get_response_header(message, "X-Ximian");
    if (response && response[0] == '1') {
        const char *mirror_url;
        GtkWidget *refresh_button;

        /* Success. Write this out to the file if it's not already there. */
        if (!g_file_exists("/etc/ximian/partnernet")) {
            int fd;

            rc_mkdir("/etc/ximian", 0755);
            fd = open(
                "/etc/ximian/partnernet", O_WRONLY | O_CREAT | O_TRUNC, 0600);
            write(fd, info->secret, 37);
            close(fd);
        }

        /* Get the mirror URL, which is sent back to us from the server. */
        mirror_url = soup_message_get_response_header(message, "X-MirrorURL");
        if (!mirror_url)
            g_warning("No X-MirrorURL response from server.");
        gui_prefs_set_mirror_url(mirror_url, PRIORITY_TYPE_PRIORITY);

        refresh_button = gtk_button_new_with_label(
            _("Activate Premium Services"));
        gtk_widget_show(refresh_button);
        gtk_object_set_data(
            GTK_OBJECT(html), "refresh_button", refresh_button);
        gtk_signal_connect(
            GTK_OBJECT(refresh_button), "clicked",
            GTK_SIGNAL_FUNC(enter_priority_mode), NULL);
    }

    stream = gtk_html_begin(html);
    gtk_html_write(
        html, stream, message->response.body, message->response.length);
    gtk_html_end(html, stream, GTK_HTML_STREAM_OK);

SUBMIT_CLEANUP:
    /* Reset the method since we free the old one below */
    soup_message_set_method(message, SOUP_METHOD_POST);

    g_free(info->method);
    g_free(info->cookie);
    g_free(info->secret);
    g_free(info);
} /* submit_done */

void 
rc_gui_activation_submit(GtkHTML *html, const char *form_method, const char *form_action, const char *form_encoding, gpointer data)
{
    char *method, *action, *encoding;
    GtkHTMLStream *stream;
    char *tmp;
#if 0
    char *base_url;
    char *path;
#endif
    SoupUri *url;
    SoupContext *proxy_context;
    SoupContext *context;
    SoupMessage *message;
    SubmissionInfo *info;

    /* Clearing out the HTML page below destroys the form and thus the
       method, action, and encoding. We'll copy them here and use them
       below. */
    method = g_strdup(form_method);
    action = g_strdup(form_action);
    encoding = g_strdup(form_encoding);

    /* Clear out the HTML page and tell them to wait... */
    stream = gtk_html_begin(html);
    tmp = g_strdup_printf(
        "<html><body><b>%s</b></body></html>",
        _("Please wait, loading page..."));
    gtk_html_write(html, stream, tmp, strlen(tmp));
    gtk_html_end(html, stream, GTK_HTML_STREAM_OK);

#if 0
    /* Mmmm... URL mangling... */
    /* note that this catches both http and https */
    if (!strncmp (action, "http", strlen("http"))) {
        url = soup_uri_new(action);
        g_free(url->path);
        path = g_strdup (strrchr(action, '/'));
        url->path = g_strdup_printf("%s?%s", path, encoding);
    } else {
        base_url = gtk_object_get_data(GTK_OBJECT(html), "base_url");
        url = soup_uri_new(base_url);
        path = g_strndup(url->path, (strrchr(url->path, '/') + 1) - url->path);
        g_free(url->path);
        url->path = g_strdup_printf("%s%s?%s", path, action, encoding);
        g_free(path);
    }
#else
    url = soup_uri_new(action);
    url->querystring = g_strdup_printf("%s", encoding);

    /* Assert that "action", given back by the page that handles the form
     * must always return a valid URI
     */
    g_assert(url);
#endif

    if (gui_prefs_get_proxy_url())
        proxy_context = soup_context_get(gui_prefs_get_proxy_url());
    else
        proxy_context = NULL;
    soup_set_proxy(proxy_context);

    context = soup_context_from_uri(url);
    message = soup_message_new(context, NULL);
    if (gui_prefs_get_http10_enabled())
        soup_message_set_http_version(message, SOUP_HTTP_1_0);
    soup_message_set_method(message, method);
    soup_context_unref(context);
    soup_uri_free(url);

    /* We can't free the method here, but we can free the others */
    g_free(action);
    g_free(encoding);

    info = g_new(SubmissionInfo, 1);
    info->method = method;
    info->secret = rc_get_secret(TRUE);
    info->cookie = rc_get_cookie(TRUE);
    info->html = html;

    soup_message_set_request_header(message, "X-Secret", info->secret);
    soup_message_set_request_header(message, "X-Cookie", info->cookie);
    soup_message_queue(message, submit_done, info);
} /* rc_gui_activation_submit */

void
rc_gui_enter_activation_code(void)
{
    GtkHTML *html = appinfo->exec_body;
    gboolean nocookie;
    char *activation_host;
    char *activation_url;
    char *uri;

    do {
        nocookie = g_file_exists("/etc/ximian/no-cookie") &&
            !g_file_exists("/etc/ximian/mcookie");
        
        if (!nocookie)
            break;
        else {
            gboolean retry = FALSE;

            retry = rc_gui_error_box(
                _("Red Carpet is unable to create a necessary cookie for your "
                  "machine used in authenticating you when using the Red "
                  "Carpet Premium Services because a special \"no cookie\" "
                  "file exists.  You will have to remove that file before you "
                  "can activate these services.  The file is "
                  "/etc/ximian/no-cookie."),
                RC_ERROR_RETRY);

            if (!retry)
                return;
        }
    } while (nocookie);

    rc_gui_html_sensitize_subscribe(FALSE);
    rc_gui_html_sensitize_unsubscribe(FALSE);
    rc_gui_sensitize_select_items(SELECT_NEITHER);
    rc_gui_set_refresh_item(REFRESH_SUMMARY);
    gtk_widget_set_sensitive(GET_WIDGET("summary_refresh"), FALSE);

    gtk_widget_hide(GET_WIDGET("exec_head_window"));
    gtk_widget_hide(GET_WIDGET("action_html_window"));
    rc_gui_set_page("minor_notebook", PAGE_SUMMARY);

    activation_host = getenv("RCX_ACTIVATION_URL");

    if (appinfo->rcd_host_url) {
        /* Ew. */
        activation_host = g_strdup_printf ("%s/../activation/reg-code.php",
                                           appinfo->rcd_host_url);
    }

    if (!activation_host)
        activation_host = ACTIVATION_CODE_URL;

    activation_url = g_strdup_printf(
        "%s?mid=%s&secret=%s&activationversion=%d",
        activation_host, rc_get_cookie(TRUE),
        rc_get_secret(TRUE), ACTIVATION_VERSION);

    uri = g_strdup_printf("updater:web?%s", activation_url);

    g_free(gtk_object_get_data(GTK_OBJECT(html), "base_url"));
    gtk_object_set_data(GTK_OBJECT(html), "base_url", activation_url);
    rc_gui_handle_uri(html, uri, NULL);

    g_free(uri);
} /* rc_gui_enter_activation_code */

SoupUri *
rc_gui_priority_authorized_url(const char *url)
{
    SoupUri *uri;

    uri = soup_uri_new(url);

    if (!uri) {
        /* Invalid URL was passed to us */
        return NULL;
    }

    uri->user = rc_get_cookie(FALSE);
    uri->passwd = rc_get_secret(FALSE);

    return uri;
} /* rc_gui_priority_authorized_url */

const char *
rc_gui_priority_host(void)
{
    const char *host = getenv("RCX_MAGIC");

    if (host)
        return host;

    if (appinfo->rcd_host_url)
        return appinfo->rcd_host_url;

    host = gui_prefs_get_mirror_url();

    if (host)
        return host;

    rc_debug(
        RC_DEBUG_LEVEL_WARNING, 
        "Returning default priority host.  "
        "This usually shouldn't happen\n");
    gui_prefs_set_mirror_url(
        DEFAULT_PRIORITY_HOST_ROOT, PRIORITY_TYPE_PRIORITY);
    return DEFAULT_PRIORITY_HOST_ROOT;
} /* rc_gui_priority_host */

gboolean
rc_gui_priority_is_initialized(void)
{
    if (priority_set)
        return TRUE;
    else
        return FALSE;
} /* rc_gui_priority_is_initialized */

void
rc_gui_priority(void)
{
    gboolean maybe_priority;

    maybe_priority = (rc_get_secret(FALSE) != NULL);

    gtk_widget_set_sensitive(GET_WIDGET("priority_frame"), maybe_priority);

    if (maybe_priority && gui_prefs_get_priority())
        set_priority_mode(TRUE);
    else
        set_priority_mode(FALSE);

    priority_set = TRUE;
} /* rc_gui_priority */
