/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * gui-rdf.c: RDF Parser
 *
 * The actual RDF parsing code is taken from Alan Cox's Portaloo.
 *
 * Copyright 2000, 2001 Ximian, Inc.
 * Copyright 1998 Alan Cox
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <time.h>

#include "libredcarpet.h"

#include "gui-rdf.h"

static char *
layer_find (xmlNodePtr node, char *match, char *fail)
{
    while (node) {
#ifdef RDF_DEBUG
        xmlDebugDumpNode (stdout, node, 32);
        printf("%s.\n", node->name);
#endif
        if (g_strcasecmp (node->name, match) == 0) {
            if (node->childs != NULL && node->childs->content != NULL) {
                return g_strdup(node->childs->content);
            } else {
                if (fail)
                    return g_strdup(fail);
                else
                    return NULL;
            }
        }
        node = node->next;
    }
    return g_strdup(fail);
} /* layer_find */

static char *
layer_find_url (xmlNodePtr node, char *match, char *fail)
{
    char *p = layer_find (node, match, fail);
    char *r = p;
    GString *wb;
    char *w;


    if (!p)
        return NULL;

    wb = g_string_new(NULL);

    if (*r == ' ') r++;     /* Fix UF bug */
    
    while (*r) {
        if (strncmp (r, "&amp;", 5) == 0) {
            wb = g_string_append_c(wb, '&');
            r += 5;
            continue;
        }
        if (strncmp (r, "&lt;", 4) == 0) {
            wb = g_string_append_c(wb, '<');
            r += 4;
            continue;
        }
        if (strncmp (r, "&gt;", 4) == 0) {
            wb = g_string_append_c(wb, '>');
            r += 4;
            continue;
        }
        if (*r == '"' || *r == ' ') {
            g_string_sprintfa(wb, "%%%02x", *r);
            r++;
            continue;
        }
        wb = g_string_append_c(wb, *r++);
    }

    w = wb->str;
    g_string_free(wb, FALSE);

    return w;
} /* layer_find_url */

static void 
tree_walk(xmlNodePtr root, RDFSummary *summary)
{
    xmlNodePtr walk;
    xmlNodePtr rewalk = root;
    xmlNodePtr channel = NULL;
    xmlNodePtr image = NULL;
    xmlNodePtr item[16];
    int items = 0;
    int i;
    char *t;
    
    do {
        walk = rewalk;
        rewalk = NULL;
        
        while (walk != NULL) {
#ifdef RDF_DEBUG
            printf ("%p, %s\n", walk, walk->name);
#endif
            if (strcasecmp (walk->name, "rdf") == 0) {
                rewalk = walk->childs;
                walk = walk->next;
                continue;
            }

            if (strcasecmp (walk->name, "rss") == 0){
                rewalk = walk->childs;
                walk = walk->next;
                continue;
            }

            /* This is the channel top level */
            if (strcasecmp (walk->name, "channel") == 0) {
                channel = walk;
                rewalk = channel->childs;
            } else if (strcasecmp (walk->name, "image") == 0) {
                image = walk;
            } else if (strcasecmp (walk->name, "item") == 0 && items < 16) {
                item[items++] = walk;
            }

            walk = walk->next;
        }
    } while (rewalk);
    
    if (channel == NULL) {
        rc_debug (RC_DEBUG_LEVEL_WARNING, "No channel definition.\n");
        return;
    }
    
    t = layer_find (channel->childs, "title", "No title");

    if (summary->title)
        g_free (summary->title);
    summary->title = t;
    
    if (image) {
        char *icon;
        
        icon = layer_find_url (image->childs, "url", "apple-red.png");
        if (summary->icon)
            g_free (summary->icon);
        summary->icon = icon;
    }

    for (i = 0; i < items; i++) {
        RDFItem *rdfitem;

        char *title = layer_find (item[i]->childs, "title", "No information");
        char *location = layer_find_url (item[i]->childs, "link", "");
        char *icon = layer_find_url (item[i]->childs, "icon", NULL);
        char *summary_text = layer_find (item[i]->childs, "summary", NULL);
        char *channel = layer_find (item[i]->childs, "channel", NULL);
        char *date = layer_find (item[i]->childs, "date", NULL);
        char *datestr = NULL;
        time_t date_int = 0;

        if (!location || !location[0])
            continue;

        if (date) {
            int slen;
            date_int = (time_t) strtol (date, NULL, 10);
            datestr = g_strdup (ctime (&date_int));
            slen = strlen(datestr);
            if (slen > 2 && datestr[slen - 2] == '\n') {
                datestr[slen - 2] = 0;
            }
            g_free (date);
            date = NULL;
        }

        rdfitem = g_new0(RDFItem, 1);
        rdfitem->title = title;
        rdfitem->location = location;
        rdfitem->icon = icon;
        rdfitem->summary = summary_text;
        rdfitem->channel = channel;
        rdfitem->datestr = datestr;
        rdfitem->date = date_int;

        summary->items = g_slist_append(summary->items, rdfitem);
    }
} /* tree_walk */

static gint
sort_by_date (gconstpointer a, gconstpointer b)
{
    RDFItem *one, *two;

    one = (RDFItem *) a;
    two = (RDFItem *) b;

    if (one->date < two->date) {
        return 1;
    } else if (one->date > two->date) {
        return -1;
    } else {
        return 0;
    }
}

RDFSummary *
rc_gui_rdf_parse(GByteArray *data)
{
    xmlDoc *doc;
    RDFSummary *summary;

    g_return_val_if_fail(data, NULL);

    data = g_byte_array_append(data, "\0", 1);
    doc = xmlParseMemory(data->data, data->len - 1);

    if (!doc)
        return NULL;

    summary = g_new0(RDFSummary, 1);

    tree_walk(doc->root, summary);

    summary->items = g_slist_sort (summary->items,
                                   (GCompareFunc) sort_by_date);

    return summary;
} /* rc_gui_rdf_parse */

